/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.annotation.valueprovider;

import static org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionEntry.ActionScope.ALL_VISIBLE_ITEMS;
import static org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionEntry.ActionScope.SINGLE_ITEM;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.BiConsumer;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EEnum;
import org.eclipse.emf.ecore.EFactory;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.fortiss.tooling.base.ToolingBaseActivator;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider.AnnotationActionEntry.ActionScope;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Base class for {@link IAnnotationValueProvider}s.
 * 
 * @author diewald, barner
 */
public abstract class ValueProviderBase<T extends IAnnotatedSpecification>
		implements IAnnotationValueProvider<T> {

	/**
	 * {@link EClass} of {@link IAnnotatedSpecification} that contains the respective structural
	 * feature and that is managed by this {@link IAnnotationValueProvider}.
	 */
	protected final EClass annotatedSpecificationEClass;

	/**
	 * Fixed input choice to be offered in a combo box. The semantics of the map value is as
	 * follows:
	 * <ul>
	 * <li>{@code fixedInputChoice == null} -> do not use fixed values, provide text input
	 * cell.</li>
	 * <li>{@code !fixedInputChoice.isEmpty()} -> provide a fixed set of choices in an input box.
	 * <li>{@code fixedInputChoice.isEmpty()} -> dynamically manage the set of provided values,
	 * i.e., the combo box is editable and enables the user to add new values and remove existing
	 * choices (by removing all instances of them).</li>
	 * <li>For enum types, the set of inputs is automatically initialized from the underlying
	 * {@link EEnum} type definition.</li>
	 * </ul>
	 */
	private List<String> fixedInputChoice;

	/**
	 * Current input choice, computed according to the initialization of {@link #fixedInputChoice}.
	 */
	private Collection<String> currentInputChoice;

	/**
	 * Last copied "annotation instance" (annotation + modelElement; unique)
	 */
	private AnnotationInstSpec<T> copiedAnnotation;

	/**
	 * Constructs a new {@link IAnnotationValueProvider} for an annotation specified in a given
	 * {@link EClass}.
	 */
	public ValueProviderBase(EClass annotatedSpecificationEClass) {
		this.annotatedSpecificationEClass = annotatedSpecificationEClass;
	}

	/** {@inheritDoc} */
	@Override
	public boolean setAnnotationName(String name, T specification) {
		// Renaming this annotation is not supported by default.
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isHidden(T specification) {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public void setAnnotationValue(String value, T specification) throws Exception {
		throw new Exception(
				"setAnnotationValue(String) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public <U> void setAnnotationValue(U value, T specification) throws Exception {
		throw new Exception(
				"setAnnotationValue(U) is not supported / has not been implemented for annotations of type " +
						specification.getClass());
	}

	/** {@inheritDoc} */
	@Override
	public List<Class<? extends EObject>> excludeModelElementTypeFromAnnotatedSpecification() {
		return null;
	}

	/**
	 * Add decorations to an annotation when its instance is newly created by
	 * {@link #getAnnotatedSpecificationForModelElement(IModelElement)}.
	 * 
	 * @param specification
	 *            annotation to be decorated
	 */
	protected void decorateAnnotationSpecification(T specification) throws Exception {
		// No decoration
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public T getAnnotatedSpecificationForModelElement(IModelElement element) throws Exception {

		Class<IAnnotatedSpecification> instanceClass =
				(Class<IAnnotatedSpecification>)annotatedSpecificationEClass.getInstanceClass();
		T rval = (T)pickFirstInstanceOf(instanceClass, element.getSpecifications());

		if(rval != null) {
			return rval;
		}

		// Create annotation
		EFactory eFactory = annotatedSpecificationEClass.getEPackage().getEFactoryInstance();
		T specification = (T)eFactory.create(annotatedSpecificationEClass);

		decorateAnnotationSpecification(specification);

		// Hook specification to model element
		element.addSpecification(specification);

		return specification;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<T> getAnnotationClazz() {
		return (Class<T>)annotatedSpecificationEClass.getInstanceClass();
	}

	/** Returns the {@link #fixedInputChoice}. */
	public List<String> getFixedInputChoice() {
		return fixedInputChoice;
	}

	/** Sets the {@link #fixedInputChoice}. */
	public void setFixedInputChoice(List<String> fixedInputChoice) {
		this.fixedInputChoice = fixedInputChoice;
	}

	/** Returns the {@link #currentInputChoice}. */
	public Collection<String> getCurrentInputChoice() {
		return currentInputChoice;
	}

	/** Sets the {@link #currentInputChoice}. */
	public void setCurrentInputChoice(Collection<String> currentInputChoice) {
		this.currentInputChoice = currentInputChoice;
	}

	/**
	 * Updates the current input choice, and returns the actual value. Called from
	 * {@link #setAnnotationValue(String, IAnnotatedSpecification)}. In particular, this
	 * method handles the removal of input choices from if these are dynamically managed (i.e.,
	 * based on the accumulated user input).
	 * 
	 * @param newValue
	 *            New value is that is about to be set in
	 *            {@link #setAnnotationValue(String, IAnnotatedSpecification)} and that is required
	 *            to update the input choice.
	 * @param specification
	 *            {@link IAnnotatedSpecification} for which the input choice is to be updated
	 * 
	 * @return Currently selected choice.
	 */
	protected String determineCurrentAnnotationValueAndUpdateInputChoice(String newValue,
			T specification) throws Exception {

		boolean noValue = newValue == null || newValue.isEmpty();
		if(noValue && currentInputChoice != null && !currentInputChoice.isEmpty()) {

			// Only one entry left: do not remove
			if(currentInputChoice.size() == 1) {
				return currentInputChoice.iterator().next();
			}

			// 1st entry is to be removed: Choose 2nd entry as new value
			String currentValue = getAnnotationValue(specification);

			Iterator<String> iter = currentInputChoice.iterator();
			if(iter.next().equals(currentValue)) {
				// Skip one entry
				String rval = iter.next();
				currentInputChoice.remove(currentValue);
				return rval;
			}

			// Otherwise, return the entry just before the entry to be removed.
			String rval = null;

			for(String s : currentInputChoice) {
				if(s != null && s.equals(currentValue)) {
					currentInputChoice.remove(currentValue);
					return rval;
				}
				rval = s;
			}

			return rval;
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public AnnotationActions<T> getContextMenuEntries(T specification) {
		AnnotationActions<T> retList = new AnnotationActions<>();

		EStructuralFeature eType = null;
		if(getEStructuralFeatureDescriptor() != null) {
			eType = getEStructuralFeatureDescriptor().getEStructuralFeature(specification);
		}
		boolean isManyRef = eType instanceof EReference && ((EReference)eType).isMany();

		retList.add(createCopyAction());
		retList.addAll(createPasteActions(isManyRef, false));
		if(isManyRef) {
			retList.addAll(createPasteActions(isManyRef, true));
		}

		return retList;
	}

	/** Copy a selected annotation from the table. */
	private AnnotationAction<T> createCopyAction() {
		BiConsumer<AnnotationInstSpec<T>, AnnotationActionParameters<T>> copyAction =
				(AnnotationInstSpec<T> spec, AnnotationActionParameters<T> parameters) -> {
					copiedAnnotation = spec;
				};
		AnnotationActionEntry copyDescriptor =
				new AnnotationActionEntry("Copy", ActionScope.SINGLE_ITEM, false);

		return new AnnotationAction<T>(copyDescriptor, copyAction);
	}

	/** Paste a selected annotation from the table. */
	private List<AnnotationAction<T>> createPasteActions(boolean isManyRef, boolean doReplace) {
		List<AnnotationAction<T>> pasteActions = new ArrayList<>();

		BiConsumer<AnnotationInstSpec<T>, AnnotationActionParameters<T>> pasteAction =
				(AnnotationInstSpec<T> spec, AnnotationActionParameters<T> parameters) -> {
					if(copiedAnnotation == null) {
						// Skip pasting if no annotation was copied.
						return;
					}

					AnnotationEntry annEntry = copiedAnnotation.getEntry();
					Object copiedValue = annEntry.getSpecificationValue(
							copiedAnnotation.getAnnotatedSpecificationType());

					T pasteAnnotation = parameters.getAnnotation();

					if(pasteAnnotation != null) {
						try {
							pasteValue(isManyRef, doReplace, copiedValue, pasteAnnotation);
						} catch(Exception e) {
							error(ToolingBaseActivator.getDefault(),
									"Could not set the value of the annotation " +
											pasteAnnotation.getName() +
											" when trying to propagate the value form the selected annotation.",
									e);
						}
					}
				};

		String pasteSuffix = !isManyRef ? "" : (doReplace ? " (replace)" : " (append)");

		AnnotationActionEntry pasteDescriptor =
				new AnnotationActionEntry("Paste" + pasteSuffix, SINGLE_ITEM, true);
		pasteActions.add(new AnnotationAction<T>(pasteDescriptor, pasteAction));

		AnnotationActionEntry pasteAllDescriptor =
				new AnnotationActionEntry("Paste to all" + pasteSuffix, ALL_VISIBLE_ITEMS, true);
		pasteActions.add(new AnnotationAction<T>(pasteAllDescriptor, pasteAction));

		return pasteActions;
	}

	/**
	 * Pastes a {@code copiedValue} to the {@code pasteAnnotation}. This method handles
	 * multi-reference lists; supporting append and replace operations.
	 */
	// NOTE: We can suppress these warnings
	@SuppressWarnings({"unchecked", "rawtypes"})
	private void pasteValue(boolean isManyRef, boolean doReplace, Object copiedValue,
			T pasteAnnotation) throws Exception {
		if(isManyRef) {
			// Do not copy values that cannot be pasted the target
			EList<EObject> castedList = new BasicEList((EList<EObject>)copiedValue);
			castedList.removeIf(r -> !canPaste(r, pasteAnnotation));
			copiedValue = castedList;
		} else if(copiedValue instanceof EReference) {
			if(!canPaste((EReference)copiedValue, pasteAnnotation)) {
				return;
			}
		}

		if(!isManyRef || doReplace) {
			setAnnotationValue(copiedValue, pasteAnnotation);
		} else {
			List<?> appendList = new ArrayList<>();
			appendList.addAll(getAnnotationValue(pasteAnnotation));
			appendList.addAll((List)copiedValue);
			setAnnotationValue(appendList, pasteAnnotation);
		}
	}

	/**
	 * Check whether the given {@link EObject} can be pasted to the target annotation. Used to
	 * prevent pasting of copied References that may not be set due to restrictions in the target
	 * annotation.
	 */
	private boolean canPaste(EObject refObj, T pasteAnnotation) {
		EStructuralFeatureDescriptor featurDescr = getEStructuralFeatureDescriptor();
		if(featurDescr.isAvailableObject(refObj, pasteAnnotation,
				pasteAnnotation.getSpecificationOf())) {
			return true;
		}
		return false;
	}

	/**
	 * Updates the current input map of the value provider.
	 * 
	 * @param viewerInput
	 *            Input object (provides by GUI).
	 * @param clazz
	 *            Annotation type
	 */
	@SuppressWarnings("unchecked")
	public void updateInputChoice(Object viewerInput, Class<T> clazz) {
		// Input choice is not set, or it is restricted to concrete set of values
		// (use SingleEnumAttributeValueProviderBase for EEnum types!)s
		List<String> fixedInputChoice = getFixedInputChoice();
		if(fixedInputChoice == null || !fixedInputChoice.isEmpty()) {
			setCurrentInputChoice(fixedInputChoice);
		} else {
			if(!(viewerInput instanceof Set<?>) || ((Set<?>)viewerInput).isEmpty() ||
					!(((Set<?>)viewerInput).iterator().next() instanceof AnnotationEntry)) {
				setCurrentInputChoice(null);
			} else {

				AnnotationEntry annotationEntry =
						((Set<AnnotationEntry>)viewerInput).iterator().next();

				EObject root = annotationEntry.getModelElement();
				while(root.eContainer() != null) {
					root = root.eContainer();
				}

				Set<String> choices = new TreeSet<String>();
				for(T specification : getChildrenWithType(root, clazz)) {
					Object annotationValue;
					try {
						annotationValue = getAnnotationValue(specification);
					} catch(Exception e) {
						annotationValue = null;
					}
					if(annotationValue != null) {
						choices.add(annotationValue.toString());
					}
				}
				setCurrentInputChoice(choices);
			}
		}
	}
}
